/*
 * tcpspy.c - TCP/IP connection monitor.
 *
 * This file is part of tcpspy, a TCP/IP connection monitor.
 *
 * Copyright (c) 2000 Tim J. Robbins. 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: tcpspy.c,v 1.16 2000/12/06 14:47:13 fyre Stab $
 */

#include <arpa/inet.h>
#include <assert.h>
#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#include <grp.h>
#include <limits.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <pwd.h>
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <unistd.h>

#include "rule.h"

static char *rcsid = "$Id: tcpspy.c,v 1.16 2000/12/06 14:47:13 fyre Stab $";

/*
 * Compile-time settings you may like to change.
 *
 * CONNTABLE_BUCKETS: The number of `buckets' in the connection hashtable,
 *                    this should be a positive prime number not much greater
 *                    65536. If you have enough memory, use 65537.
 */
#ifndef CONNTABLE_BUCKETS
#define CONNTABLE_BUCKETS 5003
#endif

static int stopped = 0, gotrule = 0, showprocs = 0;

/*
 * This structure holds everything we need to know about a connection. We
 * use unsigned long instead of (for example) uid_t, ino_t to make hashing
 * easier.
 */
typedef struct conn {
	unsigned long lcl;
	unsigned long lclp;
	unsigned long rmt;
	unsigned long rmtp;
	unsigned long uid;
	unsigned long ino;

	char exes[1024]; /* XXX this is a magic number */
	
	struct conn *next;
} conn_t;

typedef struct conntable {
	conn_t *buckets[CONNTABLE_BUCKETS];
} conntable_t;

static int ct_init (conntable_t *ct);
static int ct_hash (conn_t *c);
static int ct_add (conntable_t *ct, conn_t *c);
static int ct_find (conntable_t *ct, conn_t *c);
static int ct_read (conntable_t *ct);
static int ct_free (conntable_t *ct);
static void huntinode (ino_t i, char *buf, size_t bufsize);
static void logconn (conn_t *c, const char *action);
static void compare (conntable_t *old, conntable_t *new);
static void panic (const char *fmt, ...);
static void stopsig (int signal);

/*
 * Stop the program when certain signals are received. 
 */
static void stopsig (int signal)
{
	syslog (LOG_NOTICE, "caught signal %d, shutting down", signal);
	stopped = 1;
}

/*
 * Scream to syslog and exit.
 */
static void panic (const char *fmt, ...)
{
	va_list ap;

	va_start (ap, fmt);
	vsyslog (LOG_ERR, fmt, ap);
	va_end (ap);

	exit (EXIT_FAILURE);
}

/*
 * Initialise a connection table (hashtable).
 */
static int ct_init (conntable_t *ct)
{
	int i;
	
	assert (ct != NULL);

	for (i = 0; i < CONNTABLE_BUCKETS; i++)
		ct->buckets[i] = NULL;

	return 1;
}

/*
 * Simple hash function for connections. There is definitely room for
 * improvement here.
 */
static int ct_hash (conn_t *c)
{
	unsigned long h;

	assert (c != NULL);

	h = c->lcl ^ c->lclp ^ c->rmt ^ c->rmtp ^ c->uid ^ c->ino;

	return h % CONNTABLE_BUCKETS;
}

/*
 * Add a connection to the connection table.
 */
static int ct_add (conntable_t *ct, conn_t *c)
{
	conn_t *old, *newc;
	int bi;
	
	assert (ct != NULL);
	assert (c != NULL);

	newc = (conn_t *) malloc (sizeof (conn_t));
	if (newc == NULL) panic ("memory exhausted");	

	memcpy (newc, c, sizeof (conn_t));

	bi = ct_hash (c);
	old = ct->buckets[bi];
	ct->buckets[bi] = newc;
	ct->buckets[bi]->next = old;

	return 1;
}

/*
 * Find a connection in a table, return nonzero if found, zero otherwise.
 */
static int ct_find (conntable_t *ct, conn_t *c)
{
	conn_t *bucket;
	
	assert (ct != NULL);
	assert (c != NULL);

	bucket = ct->buckets[ct_hash (c)];
	while (bucket != NULL) {
		if ((c->lcl == bucket->lcl) && (c->lclp == bucket->lclp) && 
			(c->rmt == bucket->rmt) && (c->rmtp == bucket->rmtp) &&
			(c->uid == bucket->uid) && (c->ino == bucket->ino)) {
			return 1;
		}
		bucket = bucket->next;
	}

	return 0;
}

/*
 * Read /proc/net/tcp and add all connections to the table if connections
 * of that type are being watched.
 */
static int ct_read (conntable_t *ct)
{
	static FILE *fp = NULL;
	char buf[1024];
	conn_t c;
	
	assert (ct != NULL);

	if (fp == NULL) {
		fp = fopen ("/proc/net/tcp", "r");
		if (fp == NULL) panic ("/proc/net/tcp: %s", strerror (errno));
	}
	rewind (fp);

	if (fgets (buf, sizeof (buf), fp) == NULL) {
		panic ("/proc/net/tcp: missing header");
		return 0;
	}
	
	while (fgets (buf, sizeof (buf), fp) != NULL) {
		unsigned long st;

		if (sscanf (buf, "%*d: %lx:%lx %lx:%lx %lx %*x:%*x %*x:%*x %*x %lu %*d %lu", &c.lcl, &c.lclp, &c.rmt, &c.rmtp, &st, &c.uid, &c.ino) != 7) {
			panic ("/proc/net/tcp: bad file format");
			exit (EXIT_FAILURE);
		}
		if ((c.ino == 0) || (st != TCP_ESTABLISHED)) continue;
		if (showprocs != 0)
			huntinode ((ino_t) c.ino, c.exes, sizeof (c.exes));

		if ((gotrule != 0) && (rule_eval (c.uid, c.lcl, c.lclp, c.rmt, c.rmtp) == 0))
			continue;

		if (ct_add (ct, &c) == 0)
			return 0;
	}

	return 1;
}

/*
 * Free a connection table.
 */
static int ct_free (conntable_t *ct)
{
	int i;

	assert (ct != NULL);

	for (i = 0; i < CONNTABLE_BUCKETS; i++) {
		conn_t *c0, *c1;

		c0 = ct->buckets[i];
		while (c0 != NULL) {
			c1 = c0->next;
			free (c0);
			c0 = c1;
		}
		ct->buckets[i] = NULL;
	}

	return 1;
}

/*
 * Find names processes using an inode and put them in a buffer.
 */
static void huntinode (ino_t i, char *buf, size_t bufsize)
{
	DIR *procdir;
	struct dirent *procent;
	
	assert (buf != NULL);
	*buf = '\0';

	if ((procdir = opendir ("/proc")) == NULL)
		panic ("/proc: %s", strerror (errno));
	while ((procent = readdir (procdir)) != NULL) {
		char fdbuf[PATH_MAX];
		DIR *fddir;
		struct dirent *fdent;

		/*
		 * No test needed for "." and ".." since they don't begin
		 * with digits.
		 */
		if (! isdigit (*procent->d_name))
			continue;
		
		snprintf (fdbuf, sizeof (fdbuf), "/proc/%s/fd", 
				procent->d_name);
		
		/*
		 * We're don't always run as root, we may get EPERM here,
		 * ignore it.
		 */
		if ((fddir = opendir (fdbuf)) == NULL)
			continue;

		while ((fdent = readdir (fddir)) != NULL) {
			int len;
			char lnkbuf[PATH_MAX], lnktgt[PATH_MAX];
			char exebuf[PATH_MAX], exetgt[PATH_MAX];
			char *p;
			ino_t this_ino;
			
			if (! isdigit (*fdent->d_name))
				continue;
			snprintf (lnkbuf, sizeof (lnkbuf), "%s/%s", fdbuf, 
					fdent->d_name);
			len = readlink (lnkbuf, lnktgt, sizeof (lnktgt));
			if (len < 0)
				continue;
			lnktgt[len] = '\0';
			if (sscanf (lnktgt, "socket:[%lu]", &this_ino) != 1)
				continue;
			if (this_ino != i)
				continue;

			snprintf (exebuf, sizeof (exebuf), "/proc/%s/exe", 
					procent->d_name);
			len = readlink (exebuf, exetgt, sizeof (exetgt));
			if (len < 0)
				continue;
			exetgt[len] = '\0';

			/*
			 * Replace spaces in exe name with underscores, since
			 * spaces delimit exe names.
			 */
			for (p = exetgt; *p != '\0'; p++)
				if (*p == ' ') *p = '_';

			bufsize -= snprintf (buf, bufsize, "%.20s ", exetgt);
		}

		closedir (fddir);
	}
	closedir (procdir);

	/*
	 * Chomp off trailing space.
	 */
	buf[strlen (buf) - 1] = '\0';
}

/*
 * Log a connection/disconnected.
 */
static void logconn (conn_t *c, const char *action)
{
	struct in_addr in;
	char laddr[256], raddr[256];
	char *n;
	struct passwd *pw;
	char uidbuf[256];
	char *user;
	struct servent *se;
	char lport[256], rport[256];

	assert (c != NULL);
	assert (action != NULL);

	/*
	if ((gotrule != 0) && (rule_eval (c->uid, c->lcl, c->lclp, c->rmt, c->rmtp) == 0))
		return;
	*/

	in.s_addr = c->lcl;
	n = inet_ntoa (in);
	strncpy (laddr, n, sizeof (laddr));
	laddr[sizeof (laddr) - 1] = '\0';
	in.s_addr = c->rmt;
	n = inet_ntoa (in);
	strncpy (raddr, n, sizeof (raddr));
	raddr[sizeof (raddr) - 1] = '\0';

	snprintf (lport, sizeof (lport), "%lu", c->lclp);
	snprintf (rport, sizeof (rport), "%lu", c->rmtp);

	if ((se = getservbyport (htons (c->lclp), "tcp")) != NULL) {
		strncpy (lport, se->s_name, sizeof (lport));
		lport[sizeof (lport) - 1] = '\0';
	}
	if ((se = getservbyport (htons (c->rmtp), "tcp")) != NULL) {
		strncpy (rport, se->s_name, sizeof (rport));
		rport[sizeof (rport) - 1] = '\0';
	}

	if ((pw = getpwuid ((uid_t) c->uid)) == NULL) {
		snprintf (uidbuf, sizeof (uidbuf),
				"(uid %u)", (unsigned int)(uid_t)c->uid);
		user = uidbuf;
	} else {
		user = pw->pw_name;
	}

	if (showprocs != 0)
		syslog (LOG_NOTICE, "%s: proc %s, user %s, local %s:%s, remote"
			" %s:%s",
			action, *c->exes != '\0' ? c->exes : "(unknown)", 
			user, laddr, lport, raddr, rport);
	else
		syslog (LOG_NOTICE, "%s: user %s, local %s:%s, remote %s:%s",
				action, user, laddr, lport, raddr, rport);
}

/*
 * Compare the `old' and `new' tables, logging any differences.
 */
static void compare (conntable_t *old, conntable_t *new)
{
	int i;
	
	assert (old != NULL);
	assert (new != NULL);
	
	for (i = 0; i < CONNTABLE_BUCKETS; i++) {
		conn_t *bucket;

		bucket = old->buckets[i];
		while (bucket != NULL) {
			if (ct_find (new, bucket) == 0)
				logconn (bucket, "disconnect");	
			bucket = bucket->next;
		}
	}

	for (i = 0; i < CONNTABLE_BUCKETS; i++) {
		conn_t *bucket;

		bucket = new->buckets[i];
		while (bucket != NULL) {
			if (ct_find (old, bucket) == 0)
				logconn (bucket, "connect");
			bucket = bucket->next;
		}
	}
}

static void usage (void)
{
	fprintf (stderr, "usage: tcpspy [-d] [-p] [-e rule] [-I interval]\n");
	exit (EXIT_FAILURE);
}

int main (int argc, char *argv[])
{
	conntable_t old, new;
	unsigned long interval = 1000;
	pid_t p;
	int ch;
	uid_t dropuser = -1;
	gid_t dropgroup = -1, defgroup = -1;
	int debug = 0;

	/*
	 * Parse our arguments.
	 */
	opterr = 0;
	while ((ch = getopt (argc, argv, "dpe:I:U:G:")) != -1) {
		switch (ch) {
			case 'd':
				debug = 1;
				break;
			case 'p':
				showprocs = 1;
				break;
			case 'e':
				if (gotrule == 0) {
					rule_parse (optarg);
					gotrule = 1;
				} else {
					fprintf (stderr, "tcpspy: -e may only be specified once\n");
					exit (EXIT_FAILURE);
				}
				break;
			case 'I':
				interval = atoi (optarg);
				if (interval == 0) {
					fprintf (stderr, "tcpspy: bad interval\n");
					exit (EXIT_FAILURE);
				}
				break;
			case 'U':
				{
				struct passwd *pw;
				
				if (isdigit (*optarg)) {
					dropuser = atoi (optarg);
					pw = getpwuid (atoi (optarg));
				} else {
					if ((pw = getpwnam (optarg)) == NULL) {
						fprintf (stderr, "tcpspy: user `%s' unknown\n", optarg);
						exit (EXIT_FAILURE);
					}
					dropuser = pw->pw_uid;
				}
				
				/*
				 * Use the gid from the password file entry if
				 * possible, as a default.
				 */
				if (pw != NULL)
					defgroup = pw->pw_gid;
				else
					defgroup = (gid_t) -1;
				
				}
				break;
			case 'G':
				if (isdigit (*optarg))
					dropgroup = atoi (optarg);
				else {
					struct group *gr;

					if ((gr = getgrnam (optarg)) == NULL) {
						fprintf (stderr, "tcpspy: group `%s' unknown\n", optarg);
						exit (EXIT_FAILURE);
					}
					dropgroup = gr->gr_gid;
				}	
				break;
			default:
				usage();
		}
	}

	argc -= optind;
	argv += optind;

	/*
	 * Become an unprivileged user for safety purposes if requested.
	 */
	if ((dropgroup == (uid_t) -1) && (defgroup != (uid_t) -1))
		dropgroup = defgroup;
	if (dropgroup != (gid_t) -1) {
		if (setgid (dropgroup) < 0) {
			fprintf (stderr, "tcpspy: setgid: %s\n", strerror (errno));
			exit (EXIT_FAILURE);
		}
	}
	if (dropuser != (uid_t) -1) {
		if (setuid (dropuser) < 0) {
			fprintf (stderr, "tcpspy: setuid: %s\n", strerror (errno));
			exit (EXIT_FAILURE);
		}
	}

	/*
	 * Become a daemon..
	 */

	if (debug == 0) {
		p = fork();
		if (p < 0) {
			fprintf (stderr, "tcpspy: fork: %s\n",
					strerror (errno));
			exit (EXIT_FAILURE);
		} else if (p != 0) {
			printf ("tcpspy 1.4 started (pid %d)\n", (int) p);
			exit (EXIT_SUCCESS);
		}

		ioctl (STDIN_FILENO, TIOCNOTTY, NULL);
		close (STDIN_FILENO); 
		close (STDOUT_FILENO); 
		close (STDERR_FILENO);
	} else {
		p = getpid();
		printf ("tcpspy 1.4 started (pid %d) (debug)\n", p);
	}

	signal (SIGTERM, stopsig);
	if (debug != 0) signal (SIGINT, stopsig);

	/*
	 * Initialisation's done, let's spy on the evil lusers ;)
	 */

	openlog ("tcpspy", LOG_PID, FACILITY);
	syslog (LOG_NOTICE, "tcpspy started, monitoring connections");

	if (ct_init (&old) == 0) panic ("ct_init failed");
	if (ct_read (&old) == 0) panic ("ct_read failed");
	
	while (stopped == 0) {
		usleep (interval * 1000);
		if (ct_init (&new) == 0) panic ("ct_init failed");
		if (ct_read (&new) == 0) panic ("ct_read failed");
		compare (&old, &new);
		if (ct_free (&old) == 0) panic ("ct_free failed");
		memcpy (&old, &new, sizeof (conntable_t));
	}

	if (ct_free (&old) == 0) panic ("ct_free failed");
	closelog ();

	return EXIT_SUCCESS;
}
